/*
 *	@(#)CompressionStreamColor.java 1.3 99/08/25 17:22:23
 *
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 */

package com.sun.j3d.utils.compression ;
import javax.vecmath.* ;

/**
 * This class represents a color in a compression stream. It maintains both
 * floating-point and quantized representations.  This color may be bundled
 * with a vertex or exist separately as a global color.
 */
class CompressionStreamColor extends CompressionStreamElement {
    private int R, G, B, A ;
    private Color3f color3 ;
    private Color4f color4 ;
	    
    int colorInt[] ;

    /**
     * Create a CompressionStreamColor.
     *
     * @param stream CompressionStream associated with this element
     * @param color3 floating-point representation to be encoded
     */
    CompressionStreamColor(CompressionStream stream, Color3f color3) {
	this.color4 = null ;
	this.color3 = new Color3f(color3) ;
	this.colorInt = new int[4] ;
	stream.byteCount += 12 ;
    }

    /**
     * Create a CompressionStreamColor.
     *
     * @param stream CompressionStream associated with this element
     * @param color4 floating-point representation to be encoded
     */
    CompressionStreamColor(CompressionStream stream, Color4f color4) {
	this.color3 = null ;
	this.color4 = new Color4f(color4) ;
	this.colorInt = new int[4] ;
	stream.byteCount += 16 ;
    }

    /**
     * Quantize a floating point color to fixed point integer components of
     * the specified number of bits.  The bit length can range from a maximum
     * of 16 to a minimum of 2 bits since negative colors are not defined.<p>
     * 
     * The bit length is the total number of bits in the signed version of the
     * fixed point representation of the input color, which is assumed to
     * be normalized into the [0..1) range.  With the maximum bit length of
     * 16, 15 bits of positive colors can be represented; a bit length of 9 is
     * needed to get the 8 bit positive color size in common use.<p>
     *
     * @param stream CompressionStream associated with this element
     * @param table HuffmanTable for collecting data about the quantized
     * representation of this element
     */
    void quantize(CompressionStream stream, HuffmanTable huffmanTable) {
	// Clamp quantization.
	int quant = 
	    (stream.colorQuant < 2? 2 :
	     (stream.colorQuant > 16? 16 : stream.colorQuant)) ;

	absolute = false ;
	if (stream.firstColor || stream.colorQuantChanged) {
	    absolute = true ;
	    stream.lastColor[0] = 0 ;
	    stream.lastColor[1] = 0 ;
	    stream.lastColor[2] = 0 ;
	    stream.lastColor[3] = 0 ;
	    stream.firstColor = false ;
	    stream.colorQuantChanged = false ;
	}

	// Convert the floating point position to s.15 2's complement.
	if (color3 != null) {
	    R = (int)(color3.x * 32768.0) ;
	    G = (int)(color3.y * 32768.0) ;
	    B = (int)(color3.z * 32768.0) ;
	    A = 0 ;
	} else if (color4 != null) {
	    R = (int)(color4.x * 32768.0) ;
	    G = (int)(color4.y * 32768.0) ;
	    B = (int)(color4.z * 32768.0) ;
	    A = (int)(color4.w * 32768.0) ;
	}

	// Clamp color components.
	R = (R > 32767? 32767: (R < 0? 0: R)) ;
	G = (G > 32767? 32767: (G < 0? 0: G)) ;
	B = (B > 32767? 32767: (B < 0? 0: B)) ;
	A = (A > 32767? 32767: (A < 0? 0: A)) ;
	    
	// Compute quantized values.
	R &= quantizationMask[quant] ;
	G &= quantizationMask[quant] ;
	B &= quantizationMask[quant] ;
	A &= quantizationMask[quant] ;

	// Copy and retain absolute color for mesh buffer lookup.
	colorInt[0] = R ;
	colorInt[1] = G ;
	colorInt[2] = B ;
	colorInt[3] = A ;

	// Compute deltas.
	R -= stream.lastColor[0] ;
	G -= stream.lastColor[1] ;
	B -= stream.lastColor[2] ;
	A -= stream.lastColor[3] ;

	// Update last values.
	stream.lastColor[0] += R ;
	stream.lastColor[1] += G ;
	stream.lastColor[2] += B ;
	stream.lastColor[3] += A ;

	// Find the size of the largest component.
	length = getMaximum(getLength(R), getLength(G), getLength(B),
			    getLength(A)) ;

	// Find the maximum right shift that can be applied.  This is the
	// number of trailing zeroes common to all components.
	if (color4 != null)
	    shift = getMinimum(getShift(R), getShift(G), getShift(B),
			       getShift(A)) ;
	else
	    shift = getMinimum(getShift(R), getShift(G), getShift(B)) ;

	// Add this element to the Huffman table associated with this stream.
	huffmanTable.addColorEntry(length, shift, absolute) ;
    }

    /**
     * Output a setColor command.
     *
     * @param table HuffmanTable mapping quantized representations to
     * compressed encodings
     * @param output CommandStream for collecting compressed output
     */
    void outputCommand(HuffmanTable table, CommandStream output) {
	outputColor(table, output, CommandStream.SET_COLOR, 8) ;
    }

    /**
     * Output a color subcommand.
     *
     * @param table HuffmanTable mapping quantized representations to
     * compressed encodings
     * @param output CommandStream for collecting compressed output
     */
    void outputSubcommand(HuffmanTable table, CommandStream output) {

	outputColor(table, output, 0, 6) ;
    }

    //
    // Output the final compressed bits to the output command stream.  
    //
    private void outputColor(HuffmanTable table, CommandStream output,
			     int header, int headerLength) {
 	HuffmanNode t ;

	// Look up the Huffman token for this compression stream element.
	t = table.getColorEntry(length, shift, absolute) ;

	// Construct the color subcommand components.  The maximum length of a
	// color subcommand is 70 bits (a tag with a length of 6 followed by 4
	// components of 16 bits each).  The subcommand is therefore
	// constructed initially using just the first 3 components, with the
	// 4th component added later after the tag has been shifted into the
	// subcommand header.
	int componentLength = t.dataLength - t.shift ;
	int subcommandLength = t.tagLength + (3 * componentLength) ;

	R = (R >> t.shift) & (int)lengthMask[componentLength] ;
	G = (G >> t.shift) & (int)lengthMask[componentLength] ;
	B = (B >> t.shift) & (int)lengthMask[componentLength] ;

	long colorSubcommand = 
	    (((long)t.tag) << (3 * componentLength)) |
	    (((long)R)     << (2 * componentLength)) |
	    (((long)G)     << (1 * componentLength)) |
	    (((long)B)     << (0 * componentLength)) ;

	if (subcommandLength < 6) {
	    // The header will have some empty bits.  The Huffman tag
	    // computation will prevent this if necessary.
	    header |= (int)(colorSubcommand << (6 - subcommandLength)) ;
	    subcommandLength = 0 ;
	}
	else {
	    // Move the 1st 6 bits of the subcommand into the header.
	    header |= (int)(colorSubcommand >>> (subcommandLength - 6)) ;
	    subcommandLength -= 6 ;
	}

	// Add alpha if present.
	if (color4 != null) {
	    A = (A >> t.shift) & (int)lengthMask[componentLength] ;
	    colorSubcommand = (colorSubcommand << componentLength) | A ;
	    subcommandLength += componentLength ;
	}

	// Add the header and body to the output buffer.
	output.addCommand(header, headerLength,
			  colorSubcommand, subcommandLength)  ;
    }

    public String toString() {
	String d = absolute? "" : "delta " ;
	String c = (color4 == null? color3.toString(): color4.toString()) ;

	return
	    "color: " + c + "\n" +
	    " fixed point " + d + + R + " " + G + " " + B + "\n" +
	    " length " + length + " shift " + shift +
	    (absolute? " absolute" : " relative") ;
    }
}
